import java.util.Iterator;

public class LinearProbeMap {
  // Entry class
  public final static class Entry {
    Entry (Object key, Object value) {
      this.key = key;
      this.value = value;
      this.hash = getHash(key);
    }

    public final Object	key;
    public final Object	value;
    public final int hash;
  }
  
  static final Entry DELETED = new Entry(null, null);

  // Bob Jenkin's mix function
  public static int rehash (int c) {
    int a,b;
    a = b = 0xcafebabe;
    a -= b; a -= c; a ^= (c>>13);
    b -= c; b -= a; b ^= (a<<8);
    c -= a; c -= b; c ^= (b>>13);
    a -= b; a -= c; a ^= (c>>12);
    b -= c; b -= a; b ^= (a<<16);
    c -= a; c -= b; c ^= (b>>5);
    a -= b; a -= c; a ^= (c>>3);
    b -= c; b -= a; b ^= (a<<10);
    c -= a; c -= b; c ^= (b>>15);
    return c;
  }

  public static int stepFromHash (int hash) {
    return rehash(hash);
  }
  
  static final double MAX_FILL = 0.75;
  static final double SHRINK_EMPTY = 0.03;
  static final double DELETE_REHASH = 0.8;

  Entry[]	data;
  int		mask;
  int		allocSize;
  int		elementCount;
  int		deleteCount;
  int		growSize;
  int		shrinkSize;
  int		deleteSize;

  LinearProbeMap () {
    this(16);
  }
  
  LinearProbeMap (int initialSize) {
    elementCount = 0;
    deleteCount = 0;
    allocSize = 0;    
    int i = 0;
    while (initialSize != 0) { ++i; initialSize >>= 1; }
    if (i < 4) i = 4;
    resize(1 << (i-1));
  }

  void resize (int size) {
    final int	  _allocSize = allocSize;
    final Entry[] _data = data;

    allocSize = size;
    mask = size-1;

    growSize = (int)( (size * MAX_FILL) - 1 );
    if (growSize < 2) growSize = 2;
  
    shrinkSize = (int)(size * SHRINK_EMPTY);
    if (size <= 8) shrinkSize = 0;
  
    deleteSize = (int)( (size * DELETE_REHASH) - 1);

    data = new Entry[size];

    deleteCount = 0;

    int elementCount = 0;
    
    // reinsert the values
    if (_data != null) {
      final Entry[] data = this.data;
      for (int i = 0; i < _allocSize; ++i) {
        Entry newEntry = _data[i];
        if ((newEntry != null) && (newEntry != DELETED)) {
          final int	hash = newEntry.hash;
          int           index = hash & mask;
          final int     initialIndex = index;
          int           step = stepFromHash(hash);
          
          for (;;) {
            if (data[index] == null) {
              ++elementCount;
              data[index] = newEntry;
              break;
            }
        
            index = (index + step) & mask;
            if (index == initialIndex) { index = (index + 1) & mask; step = 1; }
          }
        }
      }
    }
    
    this.elementCount = elementCount;
  }

  static long bucket_collisions = 0;
  static long hash_collisions = 0;
  
  Entry find (Object key) {
    final Entry[] data = this.data;
    final int	  mask = this.mask;
    final int	  hash = getHash(key);
    int           index = hash & mask;
    final int     initialIndex = index;
    int           step = stepFromHash(hash);

    // linear probe
    for(;;) {
      Entry entry = data[index];
      if (entry == null) return null;
      if (entry != DELETED && entry.hash == hash && entry.key.equals(key)) return entry;
      index = (index + step) & mask;
      if (index == initialIndex) { index = (index + 1) & mask; step = 1; }
    }
  }

  static Entry[] _entries = new Entry[100000];
  
  static {
    for (int i = 0; i < _entries.length; ++i) {
      _entries[i] = new Entry("key" + i, "value" + i);
    }
  };
  
  
  public void put (Object key, Object value) {
  Entry entry = new Entry(key, value);
 //   putEntry(_entries[elementCount]);
    putEntry(entry);
  }
  
  void putEntry (Entry newEntry) {
      final Object  key = newEntry.key;
      final Entry[] data = this.data;
      final int	    mask = this.mask;
      final int	    hash = newEntry.hash;
      int           index = hash & mask;
      final int     initialIndex = index;
      int           step = stepFromHash(hash);
    
      // linear probe
      for (;;) {
        Entry entry = data[index];
      
        if (entry == null || entry == DELETED) {
          ++elementCount;
          data[index] = newEntry;
          if (elementCount >= growSize) {
            resize(allocSize * 2);
          }
          return;
        }

        if (entry.hash == hash) {
          if (entry.key.equals(key)) {
            data[index] = newEntry;
            return;
          }
  //        ++hash_collisions;
        }
        
//        ++bucket_collisions;
      
        index = (index + step) & mask;
        if (index == initialIndex) { index = (index + 1) & mask; step = 1; }
      }
  }

  public Object get (Object key) {
    Entry entry = find(key);
    return (entry != null) ? entry.value : null;
  }

  public boolean contains (Object key) {
    return find(key) != null;
  }

  public Iterator<Object> iterator () {
    return new Iterator<Object> () {
      int index = 0;
      public Object next () {
        while (index < allocSize) {
          Entry entry = data[index];
          ++index;
          if (entry != null && entry != DELETED) return entry;
        }
        throw new RuntimeException (); // look up which one later
      }
      
      public boolean hasNext () {
        while (index < allocSize) {
          Entry entry = data[index];
          if (entry != null && entry != DELETED) return true;
          ++index;
        }
        return false;
      }
      
      public void remove () {
        throw new RuntimeException();
      }
    };
  }
  
/*
template <typename K, typename T, typename H>
  inline void hash_map<K, T, H>::erase(typename hash_map<K, T, H>::Entry& entry) 
  {
    entry._set = false;
    entry._deleted = true;
    --_elements;
    ++_deleted;
    if (_elements <= shrinkSize) {
      resize(_allocated / 2);
    }
  }

  pubtemplate <typename K, typename T, typename H>
  inline void hash_map<K, T, H>::erase (K key)
  {
    Entry*final		data(_data);
    final int		mask(_mask);
    final int		hash(H::instance(key));
    int				index(hash & mask);

    for(;;) {
      Entry& entry(data[index]);
      if (!entry._valid) break;
      if (entry._key == key)	{
        if (entry._set) erase(entry);
        break;
      }
      index = (index + 1) & mask;
    }
  }
  */
  
  static int getHash (Object key) {
    return rehash(key == null ? 0 : key.hashCode());
  }
  
  public int size () {
    return elementCount;
  }
}